import os
import imp

# used for generating a unique id for each check
from md5 import md5

from types import StringType, TupleType, IntType, ListType

from Message import Message

DIRECTORIES = []

def registerRuleDirectory(dr):
    if dr not in DIRECTORIES:
        DIRECTORIES.append(dr)

class RulesList:
    """ This class knows about all the rules for an object """
    def __init__(self):
        self._rules = []
        self._rules_meta_type = {}
        print "Loading rules"
        self.populate()
        print "Sorting rules"
        self.sort()

    def populate(self):
        """ Go get the rules for the meta type """
        # here there be magic... 
        tempLst = []
        d = DIRECTORIES
        for dr in d:
            for name in os.listdir(dr):
                name = os.path.splitext(name)[0]
                if name in tempLst: continue

                tempLst.append(name)
                # ok so get the name...
                try:
                    fp, path, desc = imp.find_module(name, d)
                    m = imp.load_module(name, fp, path, desc)
                except ImportError:
                    # failed to import
                    continue

                for rule in getattr(m, 'rules'):
                    rule['unique_id'] = md5(rule['name']).hexdigest()
                    self._rules.append(rule)

    def sort(self):
        """ Sort the rules into easily getable order """
        for rule in self._rules:
            if not rule.has_key('meta_type'):
                raise ValueError, "The rule does not have a meta_type"
        
            for m in rule['meta_type']:
                # append the rule to the list
                r = self._rules_meta_type.get(m, [])
                r.append(rule)
                self._rules_meta_type[m] = r
        
    def get(self, meta_type):
        return self._rules_meta_type.get(meta_type, [])

    def meta_types(self):
        m = self._rules_meta_type.keys()
        m.sort()
        return m

class Check:
    """ This class does the checking for the object """

    def __init__(self):
        self.msgs = {}
        # each time Checker is run a new list of rules
        # is built
        self.rules = RulesList()

    def checkObjects(self, objs):
        """ Check each of the objects """
        for obj in objs:
            self.checkObject(obj)

    def getRuleList(self, obj):
        """ Return a rule list for this object """
        m = obj.meta_type
        return self.rules.get(m)
    
    def _callRule(self, rule, obj):
        """ Call the rule check, given an object,
        this can be a little tricky because if its
        a tuple or list we apply the later arguments
        to the function. 

        For example:

            foo -- calls the function foo 
            (foo) -- calls the function foo
            (foo, 'bar') -- calls the function with the arg, bar
        """
        fn = rule['function']
        if isListType(fn):
            if len(fn) > 1:
                # call the first things with
                # the other things and the object
                res = fn[0](obj, *fn[1:])
            else: 
                res = fn(obj)
        else:
            res = fn(obj)

        return res
                      
    def _parseResult(self, rule, obj, res):
        """ Parse all the possible results that
        we could get from the rule """
        if isinstance(res, IntType):
            if res == 1:
                # this means it went ok
                return 0
            
        # So lets see is this an integer
        if isinstance(res, IntType):
            # yes it is so go get the level and message
            # from the rule and add it in to our message tree
            msg = Message(rule['msg'], rule['level'])
            self.addMsg(obj, msg)
        else:
            # ok so they returned a string instead of 
            # a message, that is fine...
            # we'll use this string as the message
            # instead of the one the rule says
            if isinstance(res, StringType):
                msg = Message(res, rule['level'])
                self.addMsg(obj, msg)
            # ok so they returned a list of strings
            # as messages, this is ok too
            # we'll add
            elif isListType(res):
                # loop through each message in the
                # list and add them in
                for m in res:
                    assert(isinstance(m, StringType))
                    msg = Message(m, rule['level'])
                    self.addMsg(obj, msg)
            
    def checkObject(self, obj, ignoreList=[]):
        """ Check one object """
        rules = self.getRuleList(obj)
        msgAdd = 0

        # we'll ignore objects for which we have no checks
        if not rules: return
        print ignoreList
        for rule in rules:
            # ignore rules that are NOT in our list...
            ruleId = "%s_%s" % (rule['unique_id'], obj.meta_type) 
            if ignoreList and ruleId not in ignoreList: 
                print "ignoring", ruleId
                continue
            # call the rule with the object in question
            res = self._callRule(rule, obj)
            # will return 1 if a message has been
            # added...
            msgAdd = self._parseResult(rule, obj, res) 

        # do this so we know the object has been processed
        if not msgAdd:
            self.addMsg(obj)
    
    def addMsg(self, obj, msg=None):
        """ Adds a message """
        _id = self._id(obj)
        # so each message
        # has an id which is the object id
        # the object and the messages
        lstDct = self.msgs.get(_id, {'obj':obj, 'msg':[]})
        lstMsgs = lstDct['msg']
        # only add a message if we have one
        if msg:
            lstMsgs.append(msg)
            lstDct['msg'] = lstMsgs
        self.msgs[_id] = lstDct
    
    def getMsgs(self): 
        """ This is a convenience function for 
        the user interface, its easier to
        to parse a list of dicts in zpt """
        keys = self.msgs.keys()
        keys.sort()
        return [ self.msgs[k] for k in keys ]
    
    def getMsg(self, obj): return self.msgs.get(self._id(obj))
   
    ###################################################
    # Internal stuff
   
    def _id(self, obj): return obj.absolute_url(1)

###################################################

def isListType(item):
    """ Convenience function """
    return isinstance(item, ListType) or isinstance(item, TupleType)
